# Copyright 2015-2016 MongoDB Inc.
#
# License TBD
#
"""MongoDB Foreign Data Wrapper Multicorn Implementation.
"""

from bson import Regex
from bson.json_util import loads
from bson.objectid import ObjectId
from bson.son import SON
from multicorn import ForeignDataWrapper
from multicorn.utils import log_to_postgres as log2pg, ERROR, DEBUG
from pymongo import MongoClient, uri_parser
from pymongo.errors import OperationFailure

# A map from a connection string to a MongoClient

_CLIENTS = {}

_QUALIFIER_TO_OPERATOR_MAP = {'=': '$eq',
                              '<': '$lt',
                              '<=': '$lte',
                              '>': '$gt',
                              '>=': '$gte',
                              '<>': '$ne',
                              '~~': '$regex',
                              '~~*': '$regex',
                              '!~~': '$regex',
                              '!~~*': '$regex',
                              # Corresponds to: field1 IN (val1, val2, ...)
                              ('=', True): '$in',
                              # Corresponds to: field1 NOT IN (val1, val2, ...)
                              ('<>', False): '$nin'}

def _get_client(options):
    """Ensure that at most one MongoClient instance is created per unique
    connection string/username/password/authenticationDatabase tuple"""

    connection_string = options.get('connection_string', 'mongodb://localhost')
    username = options.get('username')
    password = options.get('password')
    auth_database = options.get('authenticationDatabase', 'admin')
    auth_mechanism = options.get('authenticationMechanism', 'DEFAULT')

    conn_options = {}
    conn_options['connection_string'] = connection_string

    if username is not None:
        conn_options['username'] = username
        conn_options['password'] = password
        conn_options['authenticationDatabase'] = auth_database
        conn_options['authenticationMechanism'] = auth_mechanism

    clients_hash_key = tuple(sorted(conn_options.items()))
    client = _CLIENTS.get(clients_hash_key)

    log2pg("Connected to MongoDB with seed list of '{0}', "
           "database '{1}', "
           "collection '{2}'"
           .format(uri_parser.parse_uri(connection_string)['nodelist'],
                   options['db'], options['collection']),
           DEBUG)

    if username is not None:
        log2pg("Authenticating as '{0}' against database '{1}' with mechanism '{2}'"
               .format(username, auth_database, auth_mechanism), DEBUG)

    if client is None:
        client = MongoClient(conn_options['connection_string'])
        if username is not None:
            client[auth_database].authenticate(username, password,
                                               mechanism=auth_mechanism)
        _CLIENTS[clients_hash_key] = client

    return client


class MongoDBFdw(ForeignDataWrapper):
    """MongoDB Foreign Data Wrapper Multicorn Implementation."""

    def __init__(self, options, columns):
        super(MongoDBFdw, self).__init__(options, columns)

        if 'db' not in options:
            log2pg("Required option 'db' is missing from FOREIGN TABLE",
                   ERROR)
        if 'collection' not in options:
            log2pg("Required option 'collection' is missing from FOREIGN "
                   "TABLE",
                   ERROR)

        self.client = _get_client(options)

        self.collection = self.client[options['db']][options['collection']]

        self.columns = columns
        self.column_formatters = self._create_column_formatters()

        pipeline_string = options.get('pipeline', '[]')

        log2pg("Configuring with columns '{0}' "
               "and aggregation pipeline '{1}'"
               .format(columns, pipeline_string),
               DEBUG)

        self.default_pipeline = loads(pipeline_string)

        self.path_keys = [(('_id',), 1)]
        try:
            self.path_keys = [(tuple(index['key']), 1)
                              for index in self.collection.list_indexes()]
        except OperationFailure:
            log2pg("Failed to list indexes for collection '{0}'"
                   .format(self.collection.name),
                   DEBUG)

        self.count = self.collection.count()

    insert = update = delete = rowid_column = lambda x, *args: \
        log2pg('Write operation is not allowed!', ERROR)

    import_schema = lambda x, *args: \
        log2pg('This operation is not yet supported!', ERROR)

    def execute(self, quals, columns, sortkeys=None):
        """Execute a query in the foreign data wrapper.

        See the base class for the method contract.

        This implementation for MongoDB uses the aggregation framework to match
        documents based on the given qualifiers and project fields based on
        the given columns.

        """

        log2pg("Executing query with qualifiers '{0}' and columns '{1}'"
               .format(quals, columns), DEBUG)

        projected = dict.fromkeys(columns, True)
        if not projected:
            projected['_id'] = True
        elif '_id' not in projected:
            projected['_id'] = False

        pipeline = []

        # add a $match for each synthetic filter field and remove that field
        # from the normal qualifiers
        pipeline.extend({'$match': loads(qual.value)} for qual in quals
                        if self._is_synthetic_filter(qual.field_name))
        synthetic_quals = [qual for qual in quals
                           if self._is_synthetic_filter(qual.field_name)]
        quals = [qual for qual in quals
                 if not self._is_synthetic_filter(qual.field_name)]

        pipeline.extend(self.default_pipeline)

        for qual in quals:
            if qual.operator in _QUALIFIER_TO_OPERATOR_MAP:
                query_filter = {qual.field_name:
                                    self._create_operator_document(qual)}
                pipeline.append({'$match': query_filter})

        if sortkeys:
            pipeline.append({"$sort": SON(((sortkey.attname,
                                            -1 if sortkey.is_reversed else 1)
                                           for sortkey in sortkeys))})


        pipeline.append({'$project': projected})

        log2pg("Executing aggregation pipeline '{0}'".format(pipeline), DEBUG)

        split_columns = [column.split('.') for column in columns]

        row_count = 0
        for cur in self.collection.aggregate(pipeline, allowDiskUse=True):
            row_count += 1
            row = {}
            for columns_index, column in enumerate(columns):
                cur_val = cur

                for field in split_columns[columns_index]:
                    if field not in cur_val:
                        break
                    cur_val = cur_val[field]
                else:
                    row[column] = cur_val

            for qual in synthetic_quals:
                row[qual.field_name] = qual.value

            # log2pg("Yielding row '{0}'".format(row), DEBUG)

            yield row

        log2pg("Completed query with row count {0}".format(row_count), DEBUG)


    def can_sort(self, sortkeys):
        """
        Method called from the planner to ask the FDW what are the sorts it can
        enforced, to avoid PostgreSQL sorting the data after retrieving all the
        rows. These sorts can come from explicit ORDER BY clauses, but also
        GROUP BY and DISTINCT clauses.

        The FDW has to inspect every sort, and respond which one are handled.
        The sorts are cumulatives. For example::

            col1 ASC
            col2 DESC

        means that the FDW must render the tuples sorted by col1 ascending and
        col2 descending.

        Args:
            pathkeys (list): A list of :class:`SortKey`
                representing all the sorts the query must enforce.

        Return:
            The list of cumulative SortKey, for which the FDW can
            enforce the sort.
        """

        log2pg('Checking sortkeys {0}'.format(sortkeys), DEBUG)

        enforced_sortkeys = [
            sortkey for sortkey in sortkeys
            if sortkey.collate is None]
        log2pg('Can sort with sortkeys {0}'.format(enforced_sortkeys), DEBUG)
        return enforced_sortkeys

    def get_path_keys(self):
        """
        Method called from the planner to add additional Path to the planner.
        By default, the planner generates an (unparameterized) path, which
        can be reasoned about like a SequentialScan, optionally filtered.

        :return:
         Returns:
            A list of tuples of the form: (key_columns, expected_rows),
            where key_columns is a tuple containing the columns on which
            the path can be used, and expected_rows is the number of rows
            this path might return for a simple lookup.
            For example::

                [(('id',), 1), (('f1', 'f2'), 1)]
        """
        return self.path_keys

    def get_rel_size(self, quals, columns):
        """
        Method called from the planner to estimate the resulting relation
        size for a scan.

        It will help the planner in deciding between different types of plans,
        according to their costs.

        Args:
            quals (list): A list of Qual instances describing the filters
                applied to this scan.
            columns (list): The list of columns that must be returned.

        Returns:
            A tuple of the form (expected_number_of_rows, avg_row_width (in bytes))
        """
        return (self.count, 1024)

    def _create_operator_document(self, qual):
        """

        :param qual: the qualifier to translate into a MongoDB operator document
        :return: the MongoDB operator document, e.g. {'$eq': 1}
        """
        mongodb_operator = _QUALIFIER_TO_OPERATOR_MAP[qual.operator]
        if mongodb_operator == '$regex':
            pattern = '^' + qual.value.replace('%', '.*').replace('_', '.') + '$'
            if qual.operator.endswith('*'):
                options = 'i'
            else:
                options = 0

            regex = Regex(pattern, options)

            if qual.operator.startswith('!'):
                doc = {'$not': regex}
            else:
                doc = {'$regex': regex}
        else:
            doc = {mongodb_operator: self._pg_to_mongodb(qual)}

        return doc

    def _is_synthetic_filter(self, field_name):
        """
        :param field_name: the name of the field
        :return: true if the field name is a synthetic field of type 'filter
        and should therefore be treated as a MongoDB query filter
        """
        return self.columns[field_name].options.get('synthetic_type') == 'filter'

    def _create_column_formatters(self):
        """
        Create a dictionary of column name to column formatters

        :param columns: dictionary of column name to multicorn.ColumnDefinition
        :return: dictionary of column name to column formatter lambda
        """
        column_formatters = {}
        for column in self.columns.values():
            if (column.base_type_name == 'character varying'
                    and column.options.get('type') == 'ObjectId'):
                column_formatters[column.column_name] = ObjectId
            else:
                column_formatters[column.column_name] = None

        return column_formatters

    def _pg_to_mongodb(self, qual):
        """
        Translate a value from a Postgres value into what's expected by MongoDB

        :param qual: A multicorn.Qual instance
        :return: the value to pass to MongoDB in the filter
        """

        column_formatter = self.column_formatters[qual.field_name]

        if column_formatter is None:
            return qual.value
        else:
            return column_formatter(qual.value)
